import View from './components/view';
import Views from './components/RouterViews';
import Link from './components/link';
import { CURRENT_VIEW, VIEWS_CACHE } from './model/ViewsModel';

export let _Vue;

export function install(Vue) {
  if (install.installed && _Vue === Vue) return;
  install.installed = true;

  _Vue = Vue;

  const isDef = v => v !== undefined;

  const registerInstance = (vm, callVal) => {
    let i = vm.$options._parentVnode;
    if (isDef(i) && isDef(i = i.data) && isDef(i = i.registerRouteInstance)) {
      i(vm, callVal);
    }
  };

  Vue.mixin({
    beforeCreate() {
      /**
       * this.$options.router 是 vue app 创建时传入
       * const app = new Vue({
       *   router
       * }).$mount('#app')
       */
      // 判断是否传入了 router
      if (isDef(this.$options.router)) {
        // 在 Vue 根级（即app）上标记 this._routerRoot = this;
        // 以利于嵌套路由容器检测嵌套层级
        this._routerRoot = this;
        this._router = this.$options.router;
        // 初始化 router
        this._router.init(this);

        // 定义动态属性，在改变时触发 RouterView 的 _render 方法
        // 通过 history/base.js 的 confirmTransition 方法中添加 nextTick 进行关联渲染
        Vue.util.defineReactive(this, '_route', this._router.history.current);
      }
      else {
        // 用于查找 router-view 组件的层次判断
        this._routerRoot = (this.$parent && this.$parent._routerRoot) || this;
      }
      // 注册 VueComponent，进行 observer 处理，把当前的router-view组件添加到对应的路由记录的instance里面
      registerInstance(this, this);
    },
    destroyed() {
      // 取消 VueComponent 的注册，router-view组件destoryed的时候将该路由记录的instance置为undefined
      registerInstance(this);
    }
  });

  Object.defineProperty(Vue.prototype, '$router', {
    get() {
      return this._routerRoot._router;
    }
  });

  Object.defineProperty(Vue.prototype, '$route', {
    get() {
      return this._routerRoot._route;
    }
  });

  Object.defineProperty(Vue.prototype, '$viewsCache', {
    get() {
      return this._routerRoot[VIEWS_CACHE];
    }
  });

  Object.defineProperty(Vue.prototype, '$view', {
    get() {
      return this._routerRoot[CURRENT_VIEW];
    }
  });

  Vue.component('RouterView', View);
  Vue.component('RouterViews', Views);
  Vue.component('RouterLink', Link);

  const strats = Vue.config.optionMergeStrategies;
  // use the same hook merging strategy for route hooks
  strats.beforeRouteEnter = strats.beforeRouteLeave = strats.beforeRouteUpdate = strats.created;
}
